﻿using System;
using System.IO;
using System.Text;
using System.Runtime.Serialization.Json;
using System.Threading.Tasks;
using WingRtmpService.ChannelManagers.Messages;
using WingRtmpService.ChannelManagers.Netease.Helper;
using WingServerCommon.Log;

namespace WingRtmpService.ChannelManagers.Netease
{
    // ReSharper disable once InconsistentNaming
    class NEChannelManager : IChannelManager
    {
        private readonly HttpClientBuilder _httpClientBuilder;

        public NEChannelManager(string appKey, string appSecret, bool forceConstantCheckSum, string constantCurrentTime, string constantCheckSum)
        {
            _httpClientBuilder = new HttpClientBuilder(appKey, appSecret, forceConstantCheckSum, constantCurrentTime, constantCheckSum);
        }

        public ICreateChannelResult CreateChannel(ChannelParameter channel)
        {
            if (channel.Parameter.ContainsKey(ChannelConstant.CreateChannelKey))
            {
                var url = "https://vcloud.163.com/app/channel/create";
                var json =
                    JsonHelper.SerializerToJsonString(
                        new CreateChannelRequest(channel.Parameter[ChannelConstant.CreateChannelKey], 0));
                string responseJson = string.Empty;
                try
                {
                    responseJson = _httpClientBuilder.HttpPost(url, json);
                    var createChannelResult = JsonHelper.DeserializerJsonResult<CreateChannelResult>(responseJson);
                    return createChannelResult;
                }
                catch (Exception e)
                {
                    Logger.WriteLineError($"Create channel error {e}, the ne response Json string : {responseJson}");
                    return null;
                }
            }

            Logger.WriteLineError("Create channel parameter is empty.");
            throw new InvalidDataException();
        }

        public Task<IDeleteChannelResult> DeleteChannelAsync(ChannelParameter channel)
        {
            return Task.Run(() => DeleteChannel(channel));
        }

        private IDeleteChannelResult DeleteChannel(ChannelParameter channel)
        {
            if (channel.Parameter.ContainsKey(ChannelConstant.DeleteChannelKey))
            {
                StopChannel(channel.Parameter[ChannelConstant.DeleteChannelKey]);
                var url = "https://vcloud.163.com/app/channel/delete";
                var json =
                    JsonHelper.SerializerToJsonString(
                        new DeleteChannelRequest(channel.Parameter[ChannelConstant.DeleteChannelKey]));
                var responseJson = _httpClientBuilder.HttpPost(url, json);
                var deleteChannelResult = JsonHelper.DeserializerJsonResult<DeleteChannelResult>(responseJson);
                return deleteChannelResult;
            }

            Logger.WriteLineError("Delete channel parameter is empty.");
            throw new InvalidDataException();
        }

        /// <summary>
        /// As Ne's support engineer says, even admin delete channel, user still can pull and push video by the address they previous get, so it may cost money.
        /// But if admin stop channel, user can not pull and push video any more.
        /// </summary>
        private void StopChannel(string channelId)
        {
            if (!string.IsNullOrWhiteSpace(channelId))
            {
                var url = "https://vcloud.163.com/app/channel/pause";
                var json =
                    JsonHelper.SerializerToJsonString(
                        new StopPushChannelRequest(channelId));
                var responseJson = _httpClientBuilder.HttpPost(url, json);
                var stopResult = JsonHelper.DeserializerJsonResult<StopChannelResult>(responseJson);
                if (stopResult.Code != (int) SendResultStatus.Success)
                {
                    responseJson = _httpClientBuilder.HttpPost(url, json);//avoid quick open and close channel don't stop channel.
                    stopResult = JsonHelper.DeserializerJsonResult<StopChannelResult>(responseJson);
                    if (stopResult.Code != (int) SendResultStatus.Success)
                    {
                        Logger.WriteLineError("Pause channel failed.");
                    }
                }

                return;
            }

            Logger.WriteLineError("Pause channel id is empty.");
            throw new InvalidDataException("Pause channel id is empty");
        }

        class JsonHelper
        {
            /// <summary>
            /// Deserializer from json result.
            /// </summary>
            /// <typeparam name="T">Type of object.</typeparam>
            /// <param name="jsonResult">Json string</param>
            /// <returns>Instance of T.</returns>
            public static T DeserializerJsonResult<T>(string jsonResult)
            {
                if (!string.IsNullOrEmpty(jsonResult))
                {
                    using (var ms = new MemoryStream(Encoding.UTF8.GetBytes(jsonResult)))
                    {                   
                        var type = typeof(T);                   
                        var serializer = new DataContractJsonSerializer(type);                                                          
                        var readObject = serializer.ReadObject(ms);
                        
                        return (T)readObject;
                    }
                }

                return default(T);
            }

            /// <summary>
            /// Serialize from instance to json string
            /// </summary>
            /// <typeparam name="T">Type of object.</typeparam>
            /// <param name="instance">Instance need to serialize.</param>
            /// <param name="useJsonFormatter"></param>
            /// <returns>The json string</returns>
            public static string SerializerToJsonString<T>(T instance, bool useJsonFormatter = true)
            {
                var type = instance.GetType();
                DataContractJsonSerializer json = new DataContractJsonSerializer(type);
                using (MemoryStream stream = new MemoryStream())
                {                
                    json.WriteObject(stream, instance);
                    string szJson = Encoding.UTF8.GetString(stream.ToArray());
                    return useJsonFormatter ? JsonFormatter.FormatJson(szJson) :
                        szJson;
                }
            }
        }

        //https://stackoverflow.com/questions/4580397/json-formatter-in-c
        class JsonFormatter
    {
        private const string IndentString = "    ";

        /// <summary>
        /// Format string to json.
        /// </summary>
        /// <param name="str">Input string.</param>
        /// <returns>Formated json string.</returns>
        public static string FormatJson(string str)
        {
            var indent = 0;
            var quoted = false;
            var sb = new StringBuilder();
            for (var i = 0; i < str.Length; i++)
            {
                var ch = str[i];
                switch (ch)
                {
                    case '{':
                    case '[':
                        sb.Append(ch);
                        if (!quoted)
                        {
                            sb.AppendLine();
                            Enumerable.Range(0, ++indent).ToList().ForEach(item => sb.Append(IndentString));
                        }
                        break;
                    case '}':
                    case ']':
                        if (!quoted)
                        {
                            sb.AppendLine();
                            Enumerable.Range(0, --indent).ToList().ForEach(item => sb.Append(IndentString));
                        }
                        sb.Append(ch);
                        break;
                    case '"':
                        sb.Append(ch);
                        bool escaped = false;
                        var index = i;
                        while (index > 0 && str[--index] == '\\')
                            escaped = !escaped;
                        if (!escaped)
                            quoted = !quoted;
                        break;
                    case ',':
                        sb.Append(ch);
                        if (!quoted)
                        {
                            sb.AppendLine();
                            Enumerable.Range(0, indent).ToList().ForEach(item => sb.Append(IndentString));
                        }
                        break;
                    case ':':
                        sb.Append(ch);
                        if (!quoted)
                            sb.Append(" ");
                        break;
                    default:
                        sb.Append(ch);
                        break;
                }
            }
            return sb.ToString();
        }
    }
    }
}
