﻿using System;
using System.IO;
using System.Numerics;
using System.Text;
using WingServerCommon.Log;
using WingServerCommon.Config;
using WingRtmpService.ChannelManagers;
using WingRtmpService.ChannelManagers.Common;
using WingRtmpService.ChannelManagers.Embedded;
using WingRtmpService.ChannelManagers.Netease.Helper;
using WingRtmpService.ChannelManagers.Messages;
using WingRtmpService.ChannelManagers.Netease;
using WingRtmpService.ChannelManagers.Tencent;
using WingRtmpService.ChannelManagers.Ali;
using WingServerCommon.Config.Parameters;

namespace WingRtmpService
{
    class ChannelOperationHelper
    {
        private static IChannelManager _channelManager;

        private static IChannelManager ChannelManager
        {
            get
            {
                if (EnvironmentConfigs.Rtmp.RTMPType.Equals("Embedded", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (_channelManager is not EmbeddedChannelManager)
                    {
                        _channelManager = new EmbeddedChannelManager();
                    }
                }
                else if (EnvironmentConfigs.Rtmp.RTMPType.Equals("Common", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (_channelManager is not CommonChannelManager)
                    {
                        var pushUrl = EnvironmentConfigs.Rtmp.RTMPPushUrl;
                        var playUrl = EnvironmentConfigs.Rtmp.RTMPPlayUrl;
                        _channelManager = new CommonChannelManager(pushUrl, playUrl);
                    }
                }
                else if (EnvironmentConfigs.Rtmp.RTMPType.Equals("Tencent", StringComparison.InvariantCultureIgnoreCase))
                {
                    if (_channelManager is not TxChannelManager)
                    {
                        _channelManager = new TxChannelManager();
                    }
                }
                Logger.WriteLineInfo($"RtmpService ChannelOperationHelper ChannelManager:{_channelManager.GetType().Name}");
                return _channelManager;
            }
        }

        static ChannelOperationHelper()
        {
        }


        public static ICreateChannelResult ApplyChannel()
        {
            var channelParameter = new ChannelParameter();
            channelParameter.SetCreateParameter(ShortGuid.NewGuid());
            var result = ChannelManager?.CreateChannel(channelParameter);
            if (result != null)
            {
                if (result.Code == (int)SendResultStatus.Success)
                {
                    return result;
                }

                Logger.WriteLineError($"Apply channel failed with result code : {result.Code}");
            }

            return null;
        }

        public static ChannelData ApplyChannel(ChannelCategory description, int retryCount = 0)
        {
            var createChannelResult = ApplyChannel();
            var channelInfo = createChannelResult?.ChannelInfo;
            if (channelInfo != null)
            {
                var channelId = channelInfo.Cid;
                var uploadUrl = channelInfo.PushUrl;
                var downloadUrl = channelInfo.RtmpPullUrl;
                var httpUrl = channelInfo.HttpPullUrl;
                var hlsUrl = channelInfo.HlsPullUrl;
                return new ChannelData(channelId, uploadUrl, downloadUrl, description, httpUrl, hlsUrl);
            }

            if (retryCount < 3)
            {
                retryCount++;
                return ApplyChannel(description, retryCount);
            }

            return null;
        }

        public static async void CloseChannel(string cid)
        {
            try
            {
                if (string.IsNullOrWhiteSpace(cid) || ChannelManager == null)
                {
                    return;
                }
                var channelParameter = new ChannelParameter();
                channelParameter.SetDeleteParameter(cid);
                var deleteResult = await ChannelManager.DeleteChannelAsync(channelParameter).ConfigureAwait(false);
                if (deleteResult.Code != (int)(SendResultStatus.Success))
                {
                    Logger.WriteLineError($"Channel {cid} close failed in cloud.");
                }
            }
            catch (Exception ex)
            {
                Logger.WriteLineError($"{ex}");
            }
        }

        //http://blog.csdn.net/hankersyan/article/details/51280361
        private class ShortGuid
        {
            // I choose the base 36 because it generates little bit longer than 62.  
            // base36 vs. base62 vs. guid  
            // 25-26 vs. 22-23 vs. 36  
            private static readonly char[] Base36 = {
                '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
                'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
                'u', 'v', 'w', 'x', 'y', 'z'
            };

            private static string ToBase36String(byte[] toConvert, bool bigEndian = false)
            {
                if (bigEndian) Array.Reverse(toConvert); // !BitConverter.IsLittleEndian might be an alternative  
                var dividend = new BigInteger(toConvert);
                var builder = new StringBuilder();
                while (dividend != 0)
                {
                    dividend = BigInteger.DivRem(dividend, 36, out var remainder);
                    builder.Insert(0, Base36[Math.Abs((int)remainder)]);
                }
                return builder.ToString();
            }

            /// <summary>
            /// Generate a 26 byte guid.
            /// </summary>
            /// <returns>26 byte guid.</returns>
            public static string NewGuid()
            {
                return $"v{ToBase36String(Guid.NewGuid().ToByteArray())}";
            }
        }
    }
}
