﻿using System;
using System.ComponentModel;
using System.IO;

namespace WingRtmpService.EmbeddedRtmpServer.IO.AMF3
{
    [TypeConverter(typeof(ByteArrayConverter))]
    [Serializable]
    [SerializedName("flex.messaging.io.ByteArray")]
    public class ByteArray
    {

        private DataInput _dataInput;
        private DataOutput _dataOutput;
        private MemoryStream _memoryStream;
        private ObjectEncoding _objectEncoding = ObjectEncoding.Amf3;
        private SerializationContext _serializationContext;

        public uint Length => (uint)_memoryStream.Length;

        public uint Position
        {
            get => (uint)_memoryStream.Position;
            set => _memoryStream.Position = value;
        }

        public uint BytesAvailable => Length - Position;

        internal MemoryStream MemoryStream => _memoryStream;

        public ObjectEncoding ObjectEncoding
        {
            get => _objectEncoding;
            set
            {
                _objectEncoding = value;
                _dataInput.ObjectEncoding = value;
                _dataOutput.ObjectEncoding = value;
            }
        }

        public ByteArray()
        {
            _memoryStream = new MemoryStream();
            ReloadStreams();
        }

        public ByteArray(SerializationContext serializationContext) : this()
        {
            _serializationContext = serializationContext;
        }

        public ByteArray(MemoryStream ms, SerializationContext serializationContext)
        {
            _serializationContext = serializationContext;
            _memoryStream = ms;
            ReloadStreams();
        }

        public ByteArray(byte[] buffer, SerializationContext serializationContext)
        {
            _serializationContext = serializationContext;
            _memoryStream = new MemoryStream(buffer);
            ReloadStreams();
        }

        private void ReloadStreams()
        {
            _dataOutput = new DataOutput(new AmfWriter(_memoryStream, _serializationContext, _objectEncoding));
            _dataInput = new DataInput(new AmfReader(_memoryStream, _serializationContext));
        }

        // The data array backing this ByteArray
        public byte[] GetBuffer()
        {
            return _memoryStream.GetBuffer();
        }

        // Returns a byte[] of the current ByteArray from start to end, without regard for the current stream position
        public byte[] ToArray()
        {
            return _memoryStream.ToArray();
        }

        #region IDataInput Members

        public bool ReadBoolean()
        {
            return _dataInput.ReadBoolean();
        }

        public byte ReadByte()
        {
            return _dataInput.ReadByte();
        }

        public byte[] ReadBytes(int count)
        {
            return _dataInput.ReadBytes(count);
        }

        public double ReadDouble()
        {
            return _dataInput.ReadDouble();
        }

        public float ReadFloat()
        {
            return _dataInput.ReadFloat();
        }

        public short ReadInt16()
        {
            return _dataInput.ReadInt16();
        }

        public int ReadInt32()
        {
            return _dataInput.ReadInt32();
        }

        public object ReadObject()
        {
            return _dataInput.ReadObject();
        }

        public ushort ReadUInt16()
        {
            return _dataInput.ReadUInt16();
        }

        public int ReadUInt24()
        {
            return _dataInput.ReadUInt24();
        }

        public uint ReadUInt32()
        {
            return _dataInput.ReadUInt32();
        }

        public string ReadUtf()
        {
            return _dataInput.ReadUtf();
        }

        public string ReadUtf(int length)
        {
            return _dataInput.ReadUtf(length);
        }

        #endregion

        #region IDataOutput members

        public void WriteBoolean(bool value)
        {
            _dataOutput.WriteBoolean(value);
        }

        public void WriteByte(byte value)
        {
            _dataOutput.WriteByte(value);
        }

        public void WriteBytes(byte[] buffer)
        {
            _dataOutput.WriteBytes(buffer);
        }

        public void WriteDouble(double value)
        {
            _dataOutput.WriteDouble(value);
        }

        public void WriteFloat(float value)
        {
            _dataOutput.WriteFloat(value);
        }

        public void WriteInt16(short value)
        {
            _dataOutput.WriteInt16(value);
        }

        public void WriteInt32(int value)
        {
            _dataOutput.WriteInt32(value);
        }

        public void WriteObject(object value)
        {
            _dataOutput.WriteObject(value);
        }

        public void WriteUInt16(ushort value)
        {
            _dataOutput.WriteUInt16(value);
        }

        public void WriteUInt24(int value)
        {
            _dataOutput.WriteUInt24(value);
        }

        public void WriteUInt32(uint value)
        {
            _dataOutput.WriteUInt32(value);
        }

        public void WriteUtf(string value)
        {
            _dataOutput.WriteUtf(value);
        }

        public void WriteUtfBytes(string value)
        {
            _dataOutput.WriteUtfBytes(value);
        }

        #endregion
    }

    public class ByteArrayConverter : TypeConverter
    {
        public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
        {
            if (destinationType == typeof(byte[]))
                return true;
            return base.CanConvertTo(context, destinationType);
        }

        public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType)
        {
            if (value == null)
                throw new ArgumentNullException();
            if (destinationType == typeof(byte[]))
            {
                return ((ByteArray)value).MemoryStream.ToArray();
            }
            return base.ConvertTo(context, culture, value, destinationType);
        }
    }
}
