﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Dynamic;
using System.Globalization;
using System.Linq;

namespace WingRtmpService.EmbeddedRtmpServer.IO
{
    [Serializable]
    [TypeConverter(typeof(AsObjectConverter))]
    public class AsObject : DynamicObject, IDictionary<string, object>
    {
        private readonly Dictionary<string, object> _underlying;

        public string TypeName { get; set; }

        public bool IsTyped => !string.IsNullOrEmpty(TypeName);

        public AsObject()
        {
            _underlying = new Dictionary<string, object>();
        }

        public AsObject(string typeName) : this()
        {
            TypeName = typeName;
        }

        public AsObject(Dictionary<string, object> dictionary) : this()
        {
            _underlying = new Dictionary<string, object>(dictionary);
        }

        public override IEnumerable<string> GetDynamicMemberNames() => _underlying.Keys;
        public override bool TryGetMember(GetMemberBinder binder, out object result) => _underlying.TryGetValue(binder.Name, out result);
        public override bool TryDeleteMember(DeleteMemberBinder binder) => _underlying.Remove(binder.Name);
        public override bool TrySetMember(SetMemberBinder binder, object value) { _underlying[binder.Name] = value; return true; }

        #region IDictionary<> members

        public int Count => _underlying.Count;
        public bool IsReadOnly => ((IDictionary<string, object>)_underlying).IsReadOnly;
        public ICollection<string> Keys => _underlying.Keys;
        public ICollection<object> Values => _underlying.Values;

        public void Add(KeyValuePair<string, object> item)
        {
            ((IDictionary<string, object>)_underlying).Add(item);
        }

        public void Clear()
        {
            _underlying.Clear();
        }

        public bool Contains(KeyValuePair<string, object> item)
        {
            return _underlying.Contains(item);
        }

        public void CopyTo(KeyValuePair<string, object>[] array, int arrayIndex)
        {
            ((IDictionary<string, object>)_underlying).CopyTo(array, arrayIndex);
        }

        public bool Remove(KeyValuePair<string, object> item)
        {
            return ((IDictionary<string, object>)_underlying).Remove(item);
        }

        public void Add(string key, object value)
        {
            _underlying.Add(key, value);
        }

        public bool ContainsKey(string key)
        {
            return _underlying.ContainsKey(key);
        }

        public object this[string key]
        {
            get => _underlying[key];
            set => _underlying[key] = value;
        }

        public bool Remove(string key)
        {
            return _underlying.Remove(key);
        }

        public bool TryGetValue(string key, out object value)
        {
            return _underlying.TryGetValue(key, out value);
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        public IEnumerator<KeyValuePair<string, object>> GetEnumerator()
        {
            return _underlying.GetEnumerator();
        }

        #endregion
    }

    public class AsObjectConverter : TypeConverter
    {
        private readonly SerializationContext _serializationContext;

        public static SerializationContext DefaultSerializationContext = new SerializationContext();

        public AsObjectConverter()
        {
            _serializationContext = DefaultSerializationContext;
        }

        public AsObjectConverter(SerializationContext serializationContext)
        {
            _serializationContext = serializationContext;
        }

        public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
        {
            if (destinationType.IsValueType || destinationType.IsEnum || destinationType.IsArray || destinationType.IsAbstract || destinationType.IsInterface)
                return false;
            return true;
        }

        public override object ConvertTo(ITypeDescriptorContext context, CultureInfo culture, object value, Type destinationType)
        {
            var instance = destinationType.CreateInstance();
            var classDescription = _serializationContext.GetClassDescription(destinationType, instance);
            if (value is IDictionary<string, object> source)
            {
                foreach (var kv in source)
                {
                    IMemberWrapper wrapper;
                    if (classDescription.TryGetMember(kv.Key, out wrapper))
                        wrapper.SetValue(instance, kv.Value);
                }
                return instance;
            }
            return base.ConvertTo(context, culture, value, destinationType);
        }
    }
}
