﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;

namespace WingRtmpService.EmbeddedRtmpServer.IO
{
    class SerializerObjectFactory
    {
        // load serializable types within this library
        private static readonly Type[] BaseSerializableTypes = typeof(SerializerObjectFactory).Assembly
            .GetTypes()
            .Where(x => x.GetCustomAttributes<SerializedNameAttribute>().Any())
            .ToArray();

        private readonly Dictionary<string, Type> _remoteToLocalNames = new Dictionary<string, Type>();
        private readonly Dictionary<string, string> _localToRemoteNames = new Dictionary<string, string>();


        public SerializerObjectFactory()
        {
            foreach (var type in BaseSerializableTypes)
                Register(type);
        }


        public object Create(Type type) => type.CreateInstance();

        public object Create(string typeName) => Create(_remoteToLocalNames[typeName]);


        // Add the type to the list of creatable types
        public void Register(Type type)
        {
            if (type.IsEnum)
                return;

            type.Register();

            var attributes = type.GetCustomAttributes<SerializedNameAttribute>(false).ToArray();
            if (attributes.Length == 0)
            {
                _remoteToLocalNames[type.FullName] = type;
            }
            else if (attributes.Count(x => x.Canonical) == 1)
            {
                foreach (var attribute in attributes)
                    RegisterAlias(type, attribute.SerializedName, attribute.Canonical);
            }
            else
            {
                throw new ArgumentException(
                    $"{type.FullName} has {attributes.Count(x => x.Canonical)} candidate names marked as canonical. only one can be canonical.",
                    nameof(type));
            }
        }

        public void RegisterAlias(Type type, string alias, bool canonical)
        {
            _remoteToLocalNames[alias] = type;
            if (canonical)
                _localToRemoteNames[type.FullName] = alias;
        }

        public string GetAlias(string typeName)
        {
            string mappedTypeName;
            return _localToRemoteNames.TryGetValue(typeName, out mappedTypeName) ? mappedTypeName : typeName;
        }

        public bool CanCreate(Type type)
        {
            return type.IsRegistered();
        }

        public bool CanCreate(string typeName)
        {
            if (_remoteToLocalNames.TryGetValue(typeName, out var type))
                return CanCreate(type);
            return false;
        }
    }
}
