﻿using System;
using System.Collections;
using System.Collections.Concurrent;
using System.Net;
using System.Threading;
using WingRtmpService.EmbeddedRtmpServer.Log;
using WingRtmpService.EmbeddedRtmpServer.Messaging.Events;

namespace WingRtmpService.EmbeddedRtmpServer.Net.Http
{
    class HttpFlvPlayer:IPlayer
    {

        private readonly HttpListenerContext _context;

        private readonly ConcurrentQueue<FlvTag> _queuedTags = new ConcurrentQueue<FlvTag>();

        // ReSharper disable once PrivateFieldCanBeConvertedToLocalVariable
        private readonly Thread _ioThread;
        private int _tagAvailable;

        public event EventHandler Closed;

        public string ChannelName { get; }

        public bool IsClosed { get; private set; }

        public HttpFlvPlayer(string channelName, HttpListenerContext context)
        {
            ChannelName = channelName;
            _context = context;
            _ioThread = new Thread(() =>
            {
                while (!IsClosed)
                {
                    try
                    {
                        WriteOnce();
                        Thread.Sleep(1);
                    }
                    catch
                    {
                        Close();
                    }
                }
            });
            _ioThread.Start();
            Logger.WriteLineDebug("Flv player polling thread started.");
        }

        private void WriteFlvHeader(bool hasVideo, bool hasAudio)
        {
            var headerBitArray = new BitArray(FlvTag.FlvHeader);
            var headerData = new byte[13];
            if (hasAudio)
            {
                headerBitArray[37] = true;
                headerBitArray.CopyTo(headerData, 0);
            }

            if (hasVideo)
            {
                headerBitArray[39] = true;
                headerBitArray.CopyTo(headerData, 0);
            }
            _context.Response.OutputStream.Write(headerData,0,headerData.Length);
        }


        private void WriteOnce()
        {
            if (Interlocked.Exchange(ref _tagAvailable, 0) == 1)
            {
                while (_queuedTags.TryDequeue(out var tag))
                {
                    WriteTag(tag);
                }
            }
        }

        private void WriteTag(FlvTag tag)
        {
           _context.Response.OutputStream.Write(tag.TagData,0,tag.TagData.Length);
        }


        private FlvTag CreateFlvTag(ByteData mediaData)
        {
            return new FlvTag(mediaData);
        }


        private void Queue(FlvTag tag)
        {
            _queuedTags.Enqueue(tag);
            Interlocked.Exchange(ref _tagAvailable, 1);
        }


        public void SendMetaData(VideoData videoConfigureRecord, AudioData audioConfigureRecord)
        {
            WriteFlvHeader(videoConfigureRecord != null, audioConfigureRecord != null);
            if (audioConfigureRecord != null)
            {
                Queue(CreateFlvTag(audioConfigureRecord));
            }

            if (videoConfigureRecord != null)
            {
                Queue(CreateFlvTag(videoConfigureRecord));
            }
        }

        public void SendAudioData(AudioData audioData)
        {
            Queue(CreateFlvTag(audioData));
        }

        public void SendVideoData(VideoData videoData)
        {
            Queue(CreateFlvTag(videoData));
        }

        public void Close()
        {
            if (!IsClosed)
            {
                _context.Response.Abort();
                IsClosed = true;
                OnClosed();
            }
        }

        private void OnClosed()
        {
            Closed?.Invoke(this, EventArgs.Empty);
        }
    }
}
