﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Text;
using System.Threading;
using WingRtmpService.EmbeddedRtmpServer.Log;
using WingRtmpService.EmbeddedRtmpServer.Messaging.Events;

namespace WingRtmpService.EmbeddedRtmpServer.Net.Http
{
    class HttpServer: IHttpExtension
    {
        private readonly Thread _listenThread;
        private readonly HttpListener _listener;

        private readonly ConcurrentDictionary<string, IPublisher> _publishers = new ConcurrentDictionary<string, IPublisher>();
        private readonly ConcurrentDictionary<string, List<IPlayer>> _players = new ConcurrentDictionary<string, List<IPlayer>>();
        private readonly ConcurrentDictionary<string, HlsProducer> _hlsProducers = new ConcurrentDictionary<string, HlsProducer>();

        private bool _isClosed;

        public HttpServer()
        {

            if (RtmpServerConfig.EnableHlsExtension)
            {
                if (!File.Exists(RtmpServerConfig.FFmpegPath))
                {
                    throw new InvalidOperationException("FFmpeg not found. The HLS extension requires ffmpeg to support.");
                }
            }
            _listener = new HttpListener();
            _listener.Prefixes.Add($"http://+:{RtmpServerConfig.HttpPort}/");
            _listener.Start();
            _listenThread = new Thread(() =>
            {
                while (!_isClosed)
                {
                    try
                    {
                        var context = _listener.GetContext();
                        ThreadPool.QueueUserWorkItem(state =>
                        {
                            var handled = false;
                            if (context.Request.HttpMethod == "GET")
                            {
                                try
                                {
                                    var uri = context.Request.Url;
                                    var extension = Path.GetExtension(uri.AbsolutePath);
                                    switch (extension)
                                    {
                                        case ".flv":
                                            if (RtmpServerConfig.EnableHttpFlvExtension)
                                            {
                                                HandleHttpFlvRequest(context);
                                                handled = true;
                                            }
                                            break;
                                        case ".m3u8":
                                        case ".ts":
                                            if (RtmpServerConfig.EnableHlsExtension)
                                            {
                                                HandleHlsRequest(context);
                                                handled = true;
                                            }
                                            break;
                                        default:
                                            Logger.WriteLineWarn($"Not supported http request {context.Request.RawUrl}.");
                                            break;
                                    }
                                }
                                catch (Exception ex)
                                {
                                    Logger.WriteLineWarn($"Handle http request error:{ex.Message}");
                                }
                            }

                            if (!handled)
                            {
                                Logger.WriteLineWarn($"Not handled http request {context.Request.RawUrl}.");
                                Write404(context);
                            }
                        });
                    }
                    catch(Exception ex)
                    {
                        Logger.WriteLineError(ex.ToString());
                    }
                }
            });
            _listenThread.Start();
            Logger.WriteLineInfo("HTTP server started.");
        }

        private void HandleHttpFlvRequest(HttpListenerContext context)
        {
            var path = context.Request.Url.AbsolutePath;
            if (path.StartsWith("/"))
            {
                path = path.Remove(0, 1);
            }

            if (path.EndsWith("/"))
            {
                path = path.Remove(path.Length - 1, 1);
            }
            context.Response.Headers.Add("Access-Control-Allow-Origin", "*");
            context.Response.ContentType = "application/octet-stream";
            context.Response.StatusCode = (int)HttpStatusCode.OK;
            var channelName = path.Replace(".flv", string.Empty);
            var player = new HttpFlvPlayer(channelName, context);
            player.Closed += (sender, args) =>
            {
                if (_players.TryGetValue(player.ChannelName, out var players))
                {
                    lock (players)
                    {
                        players.Remove(player);
                        if (players.Count == 0)
                        {
                            _players.TryRemove(player.ChannelName, out _);
                        }
                        else
                        {
                            _players.AddOrUpdate(player.ChannelName, cn => players, (cn, exist) => players);
                        }
                    }
                }
            };
            if (_publishers.TryGetValue(player.ChannelName, out var publisher))
            {
                //Send meta data
                player.SendMetaData(publisher.VideoConfigureRecord, publisher.AudioConfigureRecord);
                if (RtmpServerConfig.EnablePublishVideoCache)
                {
                    //Fast play
                    var videoCaches = publisher.GetVideoCaches();
                    foreach (var videoCache in videoCaches)
                    {
                        player.SendVideoData(videoCache);
                    }
                }
            }
            else
            {
                player.Close();
            }

            if (!player.IsClosed)
            {
                _players.AddOrUpdate(channelName, cn =>
                {
                    var players = new List<IPlayer> { player };
                    return players;
                }, (cn, players) =>
                {
                    lock (players)
                    {
                        players.Add(player);
                        return players;
                    }
                });
            }
        }

        private void HandleHlsRequest(HttpListenerContext context)
        {
            context.Response.Headers.Add("Access-Control-Allow-Origin", "*");
            context.Response.ContentType = "application/octet-stream";
            var path = context.Request.Url.AbsolutePath;
            if (path.StartsWith("/"))
            {
                path = path.Remove(0, 1);
            }

            if (path.EndsWith("/"))
            {
                path = path.Remove(path.Length - 1, 1);
            }
            var handled = false;
            var sections = path.Split('/');
            if (sections.Length == 3)
            {
                var app = sections[0];
                var stream = sections[1];
                var playFile = sections[2];
                var rtmpChannelName = $"{app}/{stream}";
                if (_publishers.ContainsKey(rtmpChannelName))
                {
                    var playFilePath = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "HLS", app, stream, playFile);
                    if (File.Exists(playFilePath))
                    {
                        var tryTimes = 0;
                        while (!handled)
                        {
                            tryTimes++;
                            if (tryTimes > 3)
                            {
                                break;
                            }
                            try
                            {
                                using (var fs = File.Open(playFilePath, FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite))
                                {
                                    context.Response.ContentLength64 = fs.Length;
                                    context.Response.StatusCode = (int)HttpStatusCode.OK;
                                    fs.CopyTo(context.Response.OutputStream);
                                    context.Response.Close();
                                }

                                handled = true;
                            }
                            catch
                            {
                                //If the file is being written, wait for 1 second for it end the writing.
                                Thread.Sleep(1000);
                            }
                        }
                    }
                }
            }

            if (!handled)
            {
                Write404(context);
            }
        }

        private void Write404(HttpListenerContext context)
        {
            var notFound = Encoding.UTF8.GetBytes("404 - resource does not exist.");
            context.Response.StatusCode = (int)HttpStatusCode.NotFound;
            context.Response.ContentType = "text/html";
            context.Response.OutputStream.Write(notFound,0, notFound.Length);
            context.Response.Close();
        }

        public void AddPublisher(IPublisher publisher)
        {
            publisher.Closed += OnPublisherClosed;

            if (RtmpServerConfig.EnableHttpFlvExtension)
            {
                //For HTTP-FLV
                publisher.VideoDataReceived += OnPublisherVideoDataReceived;
                publisher.AudioDataReceive += OnPublisherAudioDataReceive;
            }

            _publishers.AddOrUpdate(publisher.ChannelName, cn => publisher, (cn, exist) =>
            {
                if (exist != publisher)
                {
                    exist.Closed -= OnPublisherClosed;
                    exist.VideoDataReceived -= OnPublisherVideoDataReceived;
                    exist.AudioDataReceive -= OnPublisherAudioDataReceive;
                }
                return publisher;
            });

            if (RtmpServerConfig.EnableHlsExtension)
            {
                //For HLS
                var appAndStream = publisher.ChannelName.Split('/');
                if (appAndStream.Length == 2)
                {
                    var app = appAndStream[0];
                    var stream = appAndStream[1];
                    var cacheFolder = Path.Combine(RtmpServerConfig.HlsCacheFolder, app, stream);
                    var streamSource = $"rtmp://127.0.0.1:{RtmpServerConfig.RtmpPort}/{app}/{stream}";
                    var producer = new HlsProducer(streamSource, cacheFolder);
                    _hlsProducers.AddOrUpdate(publisher.ChannelName, cn => producer, (cn, exist) =>
                    {
                        if (exist != producer)
                        {
                            exist.Dispose();
                        }

                        return producer;
                    });
                }
            }
        }


        private void OnPublisherAudioDataReceive(object sender, AudioData e)
        {
            var publisher = (IPublisher)sender;
            if (RtmpServerConfig.EnableHttpFlvExtension)
            {
                if (_players.TryGetValue(publisher.ChannelName, out var players))
                {
                    lock (players)
                    {
                        foreach (var player in players)
                        {
                            try
                            {
                                player.SendAudioData(e);
                            }
                            catch (Exception ex)
                            {
                                Logger.WriteLineWarn($"Send audio data to http-flv player failed: {ex.Message}");
                            }
                        }
                    }
                }
            }
        }

        private void OnPublisherVideoDataReceived(object sender, VideoData e)
        {
            var publisher = (IPublisher)sender;
            if (RtmpServerConfig.EnableHttpFlvExtension)
            {
                if (_players.TryGetValue(publisher.ChannelName, out var players))
                {
                    lock (players)
                    {
                        foreach (var player in players)
                        {
                            try
                            {
                                player.SendVideoData(e);
                            }
                            catch (Exception ex)
                            {
                                Logger.WriteLineWarn($"Send video data to http-flv player failed: {ex.Message}");
                            }
                        }
                    }
                }
            }
        }

        private void OnPublisherClosed(object sender, EventArgs e)
        {
            var publisher = (IPublisher)sender;
            publisher.Closed -= OnPublisherClosed;
            if (RtmpServerConfig.EnableHttpFlvExtension)
            {
                publisher.VideoDataReceived -= OnPublisherVideoDataReceived;
                publisher.AudioDataReceive -= OnPublisherAudioDataReceive;
            }

            if (_publishers.TryRemove(publisher.ChannelName, out _))
            {
                if (RtmpServerConfig.EnableHttpFlvExtension)
                {
                    if (_players.TryRemove(publisher.ChannelName, out var players))
                    {
                        lock (players)
                        {
                            foreach (var player in players)
                            {
                                player.Close();
                            }

                            players.Clear();
                        }
                    }
                }

                if (RtmpServerConfig.EnableHlsExtension)
                {
                    if (_hlsProducers.TryRemove(publisher.ChannelName, out var producer))
                    {
                        producer.Dispose();
                    }
                }
            }
        }

        public void Close()
        {
            _listener.Stop();
            _isClosed = true;
            _listenThread.Join();
        }
    }
}
