﻿using System;
using System.Linq;
using System.Net.NetworkInformation;
using System.Net.Sockets;
using System.Threading;
using WingRtmpService.EmbeddedRtmpServer.IO;
using WingRtmpService.EmbeddedRtmpServer.Log;
using WingRtmpService.EmbeddedRtmpServer.Messaging;
using WingRtmpService.EmbeddedRtmpServer.Messaging.Events;
using WingRtmpService.EmbeddedRtmpServer.Messaging.Messages;

namespace WingRtmpService.EmbeddedRtmpServer.Net
{
    enum ConnectionType
    {
        Unknown,
        Publisher,
        Player
    }

    class RtmpConnection:IDisposable
    {
        public static readonly TimeSpan IoTimeout = new TimeSpan(0, 0, 3);

        private const int ProtocolMinCsid = 3;
        private const int ProtocolMaxCsid = 65599;
        private const int ControlCsid = 2;

        private static readonly IdGenerator ConnectionIdGenerator = new IdGenerator();
        private static readonly IdGenerator StreamIdGenerator = new IdGenerator();

        private readonly TcpClient _client;

        private readonly RtmpPacketWriter _writer;
        private readonly RtmpPacketReader _reader;

        private readonly Random _random = new Random();

        // ReSharper disable once PrivateFieldCanBeConvertedToLocalVariable
        private readonly Thread _ioThread;

        private string _connectedApp;

        private bool _disposed;

        public event EventHandler<ConnectEventArgs> RequestConnect;

        public event EventHandler Connected;
            
        public event EventHandler<PublishEventArgs> RequestPublish;

        public event EventHandler<PublishEventArgs> Published;

        public event EventHandler UnPublish;

        public event EventHandler<PlayEventArgs> RequestPlay;

        public event EventHandler<PlayEventArgs> Played;

        public event EventHandler Closed;

        public event EventHandler<ChannelDataReceivedEventArgs> ChannelDataReceived;

        public ConnectionType ConnectionType { get; set; } = ConnectionType.Unknown;

        public ushort ConnectionId { get; }

        public ushort StreamId { get; private set; }

        public DateTime LastPing { get; set; }

        public bool IsClosed { get; private set; }

        public bool IsConnected
        {
            get
            {
                var connectionState = IPGlobalProperties.GetIPGlobalProperties()
                                          .GetActiveTcpConnections()
                                          .SingleOrDefault(x => x.LocalEndPoint.Equals(_client.Client.LocalEndPoint)
                                                                && x.RemoteEndPoint.Equals(_client.Client.RemoteEndPoint)
                                          )?.State ?? TcpState.Unknown;
                return connectionState == TcpState.Established;
            }
        }

        public RtmpConnection(TcpClient client)
        {
            ConnectionId = ConnectionIdGenerator.GetUniqueId();
            _client = client;
            _writer = new RtmpPacketWriter(_client, ObjectEncoding.Amf0);
            _reader = new RtmpPacketReader(_client);
            _reader.EventReceived += EventReceivedCallback;
            LastPing = DateTime.UtcNow;
            if (!RtmpServerConfig.EnableGlobalPollingMode)
            {
                _ioThread = new Thread(() =>
                {
                    while (!IsClosed)
                    {
                        try
                        {
                            if (!IsConnected)
                            {
                                Logger.WriteLineWarn($"[{ConnectionType}] Connection already closed.");
                                Close();
                            }
                            else
                            {
                                var hw = WriteOnce();
                                var hr = ReadOnce();
                                if (hw || hr)
                                {
                                    LastPing = DateTime.UtcNow;
                                }
                                else
                                {
                                    if (RtmpServerConfig.AutoCleanNoOperationConnections)
                                    {
                                        if (DateTime.UtcNow - LastPing > IoTimeout)
                                        {
                                            Logger.WriteLineWarn(
                                                $"[{ConnectionType}]Connection has no writing or reading over {IoTimeout.Seconds} seconds, it will be closed.");
                                            Close();
                                        }
                                    }
                                }
                            }
                            Thread.Sleep(1);
                        }
                        catch
                        {
                            Close();
                        }
                    }
                });
                _ioThread.Start();
                Logger.WriteLineDebug("Connection polling thread started.");
            }
        }

        ~RtmpConnection()
        {
           Close();
        }

        public void SendAmf0Data(RtmpEvent e)
        {
            _writer.Queue(e, StreamId, _random.Next());
        }

        public bool WriteOnce()
        { 
           return _writer.WriteOnce();
        }

        public bool ReadOnce()
        {
           return _reader.ReadOnce();
        }

        public void Close()
        {
            if (!IsClosed)
            {
                IsClosed = true;
                OnClosed();
                DoDispose();
            }
        }


        private void EventReceivedCallback(object sender, EventReceivedEventArgs e)
        {
            switch (e.Event.MessageType)
            {
                case MessageType.UserControlMessage:
                    var m = (UserControlMessage)e.Event;
                    if (m.EventType == UserControlMessageType.PingRequest)
                    {
                        Logger.WriteLineInfo($"Receive user control message: Client Ping Request, ConnectionId={ConnectionId}");
                        WriteProtocolControlMessage(new UserControlMessage(UserControlMessageType.PingResponse, m.Values));
                    }
                    else if (m.EventType == UserControlMessageType.SetBufferLength)
                    {
                        Logger.WriteLineInfo($"Receive user control message: Set Buffer Length, ConnectionId={ConnectionId}");
                    }
                    else if (m.EventType == UserControlMessageType.PingResponse)
                    {
                        Logger.WriteLineInfo($"Receive user control message: Ping Response, ConnectionId={ConnectionId}");
                    }
                    break;

                case MessageType.DataAmf3:
                    Logger.WriteLineWarn("Receive data DataAmf3");
                    break;

                case MessageType.CommandAmf3:
                case MessageType.DataAmf0:
                case MessageType.CommandAmf0:
                    var command = (Command)e.Event;
                    var call = command.MethodCall;
                    var param = call.Parameters.Length == 1 ? call.Parameters[0] : call.Parameters;
                    switch (call.Name)
                    {
                        case "connect":
                            StreamId = StreamIdGenerator.GetUniqueId(ProtocolMinCsid, ProtocolMaxCsid);
                            HandleConnect(command);
                            Logger.WriteLineVerbose($"Client connected, StreamId={StreamId}");
                            break;
                        case "_result":
                            // unwrap Flex class, if present
                            if (param is AcknowledgeMessage)
                            {
                                Logger.WriteLineVerbose($"Receive ack message, ConnectionId={ConnectionId}");
                            }
                            break;

                        case "_error":
                            // unwrap Flex class, if present
                            if (param is ErrorMessage)
                            {
                                Logger.WriteLineVerbose($"Receive error message, ConnectionId={ConnectionId}");
                            }
                            break;
                        case "receiveAudio":
                            Logger.WriteLineVerbose($"Receive receiveAudio message, ConnectionId={ConnectionId}");
                            break;
                        case "releaseStream":
                            Logger.WriteLineVerbose($"Stream released,ConnectionId={ConnectionId}, StreamId={StreamId}");
                            break;
                        case "publish":
                            HandlePublish(command);
                            Logger.WriteLineVerbose($"Stream published({call.Name}),ConnectionId={ConnectionId}, StreamId={StreamId}");
                            break;
                        case "unpublish":
                            OnUnPublish();
                            Logger.WriteLineVerbose($"Stream un-published({call.Name}),ConnectionId={ConnectionId}, StreamId={StreamId}");
                            break;
                        case "FCpublish":
                        case "FCPublish":
                            Logger.WriteLineVerbose($"Stream published({call.Name}),ConnectionId={ConnectionId}, StreamId={StreamId}");
                            break;
                        case "FCUnpublish":
                        case "FCunPublish":
                            OnUnPublish();
                            Logger.WriteLineVerbose($"Stream un-published({call.Name}), ConnectionId={ConnectionId}, StreamId={StreamId}");
                            break;
                        case "createStream":
                            ReturnResultInvoke(null, command.InvokeId,StreamId);
                            Logger.WriteLineVerbose($"Stream created, ConnectionId={ConnectionId}, StreamId={StreamId}");
                            break;
                        case "play":
                            HandlePlay(command);
                            Logger.WriteLineVerbose($"Client request to play, ConnectionId={ConnectionId}, StreamId={StreamId}");
                            break;
                        case "deleteStream":
                            Logger.WriteLineVerbose($"Stream deleted, ConnectionId={ConnectionId}, StreamId={StreamId}");
                            break;
                        case "@setDataFrame":
                            Logger.WriteLineVerbose($"Set data frame, ConnectionId={ConnectionId}");
                            if ((string)command.ConnectionParameters != "onMetaData")
                            {
                                Logger.WriteLineError("Can only set metadata");
                                throw new InvalidOperationException("Can only set metadata");
                            }
                            break;
                        default:
                            Logger.WriteLineVerbose($"Unknown rtmp command: {call.Name}, ConnectionId={ConnectionId}");
                            break;
                    }
                    break;
                case MessageType.WindowAcknowledgementSize:
                    var msg = (WindowAcknowledgementSize)e.Event;
                    Logger.WriteLineVerbose($"Receive window acknowledgement size: {msg.Count}, ConnectionId={ConnectionId}");
                    break;
                case MessageType.Video:
                    if (e.Event is VideoData videoData)
                    {
                        OnChannelDataReceived(new ChannelDataReceivedEventArgs(ChannelType.Video, videoData));
                    }
                    break;
                case MessageType.Audio:
                    if (e.Event is AudioData audioData)
                    {
                        OnChannelDataReceived(new ChannelDataReceivedEventArgs(ChannelType.Audio, audioData));
                    }
                    break;
                case MessageType.Acknowledgement:
                    break;
                default:
                    Logger.WriteLineWarn($"Unknown message type {e.Event.MessageType}");
                    break;
            }
        }

        private void HandleConnect(Command command)
        {
            var app = ((AsObject)command.ConnectionParameters)["app"];
            if (app == null)
            {
                Logger.WriteLineError($"App value cannot be null, ConnectionId={ConnectionId}");
                Close();
            }
            else
            {
                _connectedApp = app.ToString();
                var connectArgs = new ConnectEventArgs(_connectedApp);
                if (RequestConnect != null)
                {
                    RequestConnect.Invoke(this,connectArgs);
                    string code;
                    string description;
                    if (!connectArgs.AuthSuccess)
                    {
                        code = "NetConnection.Connect.Error";
                        description = "Connection Failure.";
                    }
                    else
                    {
                        code = "NetConnection.Connect.Success";
                        description = "Connection succeeded.";
                    }

                    AsObject param = new AsObject
                    {
                        {"code", code},
                        {"description", description},
                        {"level", "status"},
                    };
                    ReturnResultInvoke(new AsObject
                    {
                        {"capabilities", 255.00},
                        {"fmsVer", "FMS/4,5,1,484"},
                        {"mode", 1.0}
                    }, command.InvokeId, param, connectArgs.AuthSuccess);
                    if (!connectArgs.AuthSuccess)
                    {
                        Logger.WriteLineError($"Auth failed, ConnectionId={ConnectionId}");
                        Close();
                    }
                    else
                    {
                        OnConnected();
                    }
                }
                else
                {
                    Logger.WriteLineWarn("No handler for connect request, it cloud be the connection already closed.");
                    Close();
                }
            }

        }

        private void HandlePublish(Command command)
        {
            string path = (string) command.MethodCall.Parameters[0];
            var publishArgs = new PublishEventArgs(_connectedApp, path);
            OnRequestPublish(publishArgs);
            if (!publishArgs.CanPublish)
            {
                Close();
            }
            else
            {
                var status = new AsObject
                {
                    {"level", "status"},
                    {"code", "NetStream.Publish.Start"},
                    {"description", "Stream is now published."},
                    {"details", path}
                };

                var callOnStatus = new InvokeAmf0
                {
                    MethodCall = new Method("onStatus", new object[] {status}),
                    InvokeId = 0,
                    ConnectionParameters = null,
                };
                callOnStatus.MethodCall.CallStatus = CallStatus.Request;
                callOnStatus.MethodCall.IsSuccess = true;

                // result.MessageType = MessageType.UserControlMessage;
                var streamBegin = new UserControlMessage(UserControlMessageType.StreamBegin, new int[] {StreamId});
                WriteProtocolControlMessage(streamBegin);
                _writer.Queue(callOnStatus, StreamId, _random.Next());
                ReturnResultInvoke(null, command.InvokeId, new object());
                OnPublished(publishArgs);
            }
        }

        private void HandlePlay(Command command)
        {
            string path = (string) command.MethodCall.Parameters[0];
            var playArgs = new PlayEventArgs(_connectedApp, path);
            OnRequestPlay(playArgs);
            if (!playArgs.CanPlay)
            {
                Close();
            }
            else
            {
                WriteProtocolControlMessage(new UserControlMessage(UserControlMessageType.StreamIsRecorded,
                    new int[] {StreamId}));
                WriteProtocolControlMessage(new UserControlMessage(UserControlMessageType.StreamBegin,
                    new int[] {StreamId}));

                var statusReset = new AsObject
                {
                    {"level", "status"},
                    {"code", "NetStream.Play.Reset"},
                    {"description", "Resetting and playing stream."},
                    {"details", path}
                };

                var callOnStatusReset = new InvokeAmf0
                {
                    MethodCall = new Method("onStatus", new object[] {statusReset}),
                    InvokeId = 0,
                    ConnectionParameters = null,
                };
                callOnStatusReset.MethodCall.CallStatus = CallStatus.Request;
                callOnStatusReset.MethodCall.IsSuccess = true;

                var statusStart = new AsObject
                {
                    {"level", "status"},
                    {"code", "NetStream.Play.Start"},
                    {"description", "Started playing."},
                    {"details", path}
                };

                var callOnStatusStart = new InvokeAmf0
                {
                    MethodCall = new Method("onStatus", new object[] {statusStart}),
                    InvokeId = 0,
                    ConnectionParameters = null,
                };
                callOnStatusStart.MethodCall.CallStatus = CallStatus.Request;
                callOnStatusStart.MethodCall.IsSuccess = true;
                _writer.Queue(callOnStatusReset, StreamId, _random.Next());
                _writer.Queue(callOnStatusStart, StreamId, _random.Next());
                OnPlayed(playArgs);
            }
        }

        private void ReturnResultInvoke(object connectParameters, int transactionId, object param, bool success = true)
        {
            var result = new InvokeAmf0
            {
                MethodCall = new Method("_result", new[] { param }),
                InvokeId = transactionId,
                ConnectionParameters = connectParameters
            };
            result.MethodCall.CallStatus = CallStatus.Result;
            result.MethodCall.IsSuccess = success;
            _writer.Queue(result, StreamId, _random.Next());
            Logger.WriteLineVerbose("Result has been queued.");
        }

        private void WriteProtocolControlMessage(RtmpEvent @event)
        {
            _writer.Queue(@event, ControlCsid, 0);
        }

        private void OnConnected()
        {
            Connected?.Invoke(this, EventArgs.Empty);
        }

        private void OnRequestPublish(PublishEventArgs e)
        {
            RequestPublish?.Invoke(this, e);
        }

        private void OnPublished(PublishEventArgs e)
        {
            Published?.Invoke(this, e);
        }

        private void OnUnPublish()
        {
            UnPublish?.Invoke(this, EventArgs.Empty);
        }

        private void OnRequestPlay(PlayEventArgs e)
        {
            RequestPlay?.Invoke(this, e);
        }

        private void OnPlayed(PlayEventArgs e)
        {
            Played?.Invoke(this, e);
        }

        private void OnClosed()
        {
            Closed?.Invoke(this, EventArgs.Empty);
        }

        private void OnChannelDataReceived(ChannelDataReceivedEventArgs e)
        {
            ChannelDataReceived?.Invoke(this, e);
        }

        private void DoDispose()
        {
            if (!_disposed)
            {
                _client.Close();
                _writer.Close();
                _reader.EventReceived -= EventReceivedCallback;
                _reader.Close();
                _disposed = true;
            }
        }

        public void Dispose()
        {
            Close();
            GC.SuppressFinalize(this);
        }
    }
}
