﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Sockets;
using System.Runtime.Serialization;
using WingRtmpService.EmbeddedRtmpServer.IO;
using WingRtmpService.EmbeddedRtmpServer.Messaging;
using WingRtmpService.EmbeddedRtmpServer.Messaging.Events;

namespace WingRtmpService.EmbeddedRtmpServer.Net
{
    // not implemented:
    //  - shared objects
    //  - amf3 data messages
    //  - multimedia packets
    public class RtmpPacketReader
    {
        private const int DefaultChunkSize = 128;

        private readonly TcpClient _client;
        private readonly AmfReader _reader;
        private readonly Dictionary<int, RtmpHeader> _rtmpHeaders = new Dictionary<int, RtmpHeader>();
        private readonly Dictionary<int, RtmpPacket> _rtmpPackets = new Dictionary<int, RtmpPacket>();

        // defined by the spec
        private int _readChunkSize = DefaultChunkSize;


        public event EventHandler<EventReceivedEventArgs> EventReceived;

        public RtmpPacketReader(TcpClient client)
        {
            _client = client;
            _reader = new AmfReader(_client.GetStream(), SerializationContext.Default);
        }

        private void OnEventReceived(EventReceivedEventArgs e) => EventReceived?.Invoke(this, e);

        public bool ReadOnce()
        {
            if (_client.Available > 0)
            {
                var header = ReadHeader();
                _rtmpHeaders[header.StreamId] = header;

                if (!_rtmpPackets.TryGetValue(header.StreamId, out var packet) || packet == null)
                {
                    packet = new RtmpPacket(header);
                    _rtmpPackets[header.StreamId] = packet;
                }

                var remainingMessageLength =
                    packet.Length + (header.Timestamp >= 0xFFFFFF ? 4 : 0) - packet.CurrentLength;
                var bytesToRead = Math.Min(remainingMessageLength, _readChunkSize);
                var bytes = _reader.ReadBytes(bytesToRead);
                packet.AddBytes(bytes);

                if (packet.IsComplete)
                {
                    _rtmpPackets.Remove(header.StreamId);

                    var rtmpEvent = ParsePacket(packet);
                    OnEventReceived(new EventReceivedEventArgs(rtmpEvent));

                    // process some kinds of packets
                    if (rtmpEvent is ChunkSize chunkSizeMessage)
                        _readChunkSize = chunkSizeMessage.Size;

                    if (rtmpEvent is Abort abortMessage)
                        _rtmpPackets.Remove(abortMessage.StreamId);
                }

                return true;
            }

            return false;
        }


        private static int GetChunkStreamId(byte chunkBasicHeaderByte, AmfReader reader)
        {
            var chunkStreamId = chunkBasicHeaderByte & 0x3F;
            switch (chunkStreamId)
            {
                // 2 bytes
                case 0:
                    return reader.ReadByte() + 64;

                // 3 bytes
                case 1:
                    return reader.ReadByte() + reader.ReadByte() * 256 + 64;

                // 1 byte
                default:
                    return chunkStreamId;
            }
        }

        private RtmpHeader ReadHeader()
        {
            // first byte of the chunk basic header
            var chunkBasicHeaderByte = _reader.ReadByte();
            var chunkStreamId = GetChunkStreamId(chunkBasicHeaderByte, _reader);
            var chunkMessageHeaderType = (ChunkMessageHeaderType)(chunkBasicHeaderByte >> 6);

            var header = new RtmpHeader()
            {
                StreamId = chunkStreamId,
                IsTimerRelative = chunkMessageHeaderType != ChunkMessageHeaderType.New
            };

            // don't need to clone if new header, as it contains all info
            if (!_rtmpHeaders.TryGetValue(chunkStreamId, out var previousHeader) && chunkMessageHeaderType != ChunkMessageHeaderType.New)
                previousHeader = header.Clone();

            switch (chunkMessageHeaderType)
            {
                // 11 bytes
                case ChunkMessageHeaderType.New:
                    header.Timestamp = _reader.ReadUInt24();
                    header.PacketLength = _reader.ReadUInt24();
                    header.MessageType = (MessageType)_reader.ReadByte();
                    header.MessageStreamId = _reader.ReadReverseInt();
                    break;

                // 7 bytes
                case ChunkMessageHeaderType.SameSource:
                    header.Timestamp = _reader.ReadUInt24();
                    header.PacketLength = _reader.ReadUInt24();
                    header.MessageType = (MessageType)_reader.ReadByte();
                    header.MessageStreamId = previousHeader.MessageStreamId;
                    break;

                // 3 bytes
                case ChunkMessageHeaderType.TimestampAdjustment:
                    header.Timestamp = _reader.ReadUInt24();
                    header.PacketLength = previousHeader.PacketLength;
                    header.MessageType = previousHeader.MessageType;
                    header.MessageStreamId = previousHeader.MessageStreamId;
                    break;

                // 0 bytes
                case ChunkMessageHeaderType.Continuation:
                    header.Timestamp = previousHeader.Timestamp;
                    header.PacketLength = previousHeader.PacketLength;
                    header.MessageType = previousHeader.MessageType;
                    header.MessageStreamId = previousHeader.MessageStreamId;
                    header.IsTimerRelative = previousHeader.IsTimerRelative;
                    break;
                default:
                    throw new SerializationException("Unexpected header type: " + (int)chunkMessageHeaderType);
            }

            // extended timestamp
            if (header.Timestamp == 0xFFFFFF)
                header.Timestamp = _reader.ReadInt32();

            return header;
        }

        private RtmpEvent ParsePacket(RtmpPacket packet, Func<AmfReader, RtmpEvent> handler)
        {
            var memoryStream = new MemoryStream(packet.Buffer, false);
            var packetReader = new AmfReader(memoryStream, _reader.SerializationContext);

            var header = packet.Header;
            var message = handler(packetReader);
            message.Header = header;
            message.Timestamp = header.Timestamp;
            return message;
        }

        private RtmpEvent ParsePacket(RtmpPacket packet)
        {
            switch (packet.Header.MessageType)
            {
                case MessageType.SetChunkSize:
                    return ParsePacket(packet, r => new ChunkSize(r.ReadInt32()));
                case MessageType.AbortMessage:
                    return ParsePacket(packet, r => new Abort(r.ReadInt32()));
                case MessageType.Acknowledgement:
                    return ParsePacket(packet, r => new Acknowledgement(r.ReadInt32()));
                case MessageType.UserControlMessage:
                    return ParsePacket(packet, r =>
                    {
                        var eventType = r.ReadUInt16();
                        var values = new List<int>();
                        while (r.Length - r.Position >= 4)
                            values.Add(r.ReadInt32());
                        return new UserControlMessage((UserControlMessageType)eventType, values.ToArray());
                    });
                case MessageType.WindowAcknowledgementSize:
                    return ParsePacket(packet, r => new WindowAcknowledgementSize(r.ReadInt32()));
                case MessageType.SetPeerBandwith:
                    return ParsePacket(packet, r => new PeerBandwidth(r.ReadInt32(), r.ReadByte()));
                case MessageType.Audio:
                    return ParsePacket(packet, r => new AudioData(packet.Buffer));
                case MessageType.Video:
                    return ParsePacket(packet, r => new VideoData(packet.Buffer));
                case MessageType.DataAmf0:
                    return ParsePacket(packet, r => ReadCommandOrData(r, new NotifyAmf0(), packet.Header));
                case MessageType.SharedObjectAmf0:
                    break;
                case MessageType.CommandAmf0:
                    return ParsePacket(packet, r => ReadCommandOrData(r, new InvokeAmf0()));
                case MessageType.DataAmf3:
                    return ParsePacket(packet, r => ReadCommandOrData(r, new NotifyAmf3()));
                case MessageType.SharedObjectAmf3:
                    break;
                case MessageType.CommandAmf3:
                    return ParsePacket(packet, r =>
                    {
                        // encoding? always seems to be zero
                        var unk1 = r.ReadByte();
                        return ReadCommandOrData(r, new InvokeAmf3());
                    });


                // aggregated messages only seem to be used in audio and video streams, so we should be OK until we need multimedia.
                // case MessageType.Aggregate:

                default:
#if DEBUG && RTMP_SHARP_DEV
                    // find out how to handle this message type.
                    System.Diagnostics.Debugger.Break();
#endif
                    break;
            }

            // skip messages we don't understand
            return null;
        }

        private static RtmpEvent ReadCommandOrData(AmfReader r, Command command, RtmpHeader header = null)
        {
            var methodName = (string)r.ReadAmf0Item();
            object temp = r.ReadAmf0Item();
            if (header != null && methodName == "@setDataFrame")
            {
                command.ConnectionParameters = temp;
            }
            else
            {
                command.InvokeId = Convert.ToInt32(temp);
                command.ConnectionParameters = r.ReadAmf0Item();
            }
            

            var parameters = new List<object>();
            while (r.DataAvailable)
                parameters.Add(r.ReadAmf0Item());

            command.MethodCall = new Method(methodName, parameters.ToArray());
            return command;
        }

        public void Close()
        {
            _reader?.Dispose();
        }
    }
}
