﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Net.Sockets;
using System.Threading;
using WingRtmpService.EmbeddedRtmpServer.IO;
using WingRtmpService.EmbeddedRtmpServer.Messaging;
using WingRtmpService.EmbeddedRtmpServer.Messaging.Events;

namespace WingRtmpService.EmbeddedRtmpServer.Net
{
    // has noop write scheduling property until we implement fair writing
    // shared objects not implemented (yet)
    class RtmpPacketWriter
    {
        // defined by the spec
        const int DefaultChunkSize = 128;

        private readonly AmfWriter _writer;
        private readonly Dictionary<int, RtmpHeader> _rtmpHeaders;
        private readonly ConcurrentQueue<RtmpPacket> _queuedPackets;
        private readonly ObjectEncoding _objectEncoding;

        private int _writeChunkSize = DefaultChunkSize;
        private int _packetAvailable;

        public RtmpPacketWriter(TcpClient client, ObjectEncoding objectEncoding)
        {
            _objectEncoding = objectEncoding;
            _writer = new AmfWriter(client.GetStream(), SerializationContext.Default, _objectEncoding);
            _rtmpHeaders = new Dictionary<int, RtmpHeader>();
            _queuedPackets = new ConcurrentQueue<RtmpPacket>();
        }

        public bool WriteOnce()
        {
            if (Interlocked.Exchange(ref _packetAvailable, 0) == 1)
            {
                while (_queuedPackets.TryDequeue(out var packet))
                    WritePacket(packet);
                return true;
            }

            return false;
        }

        private static ChunkMessageHeaderType GetMessageHeaderType(RtmpHeader header, RtmpHeader previousHeader)
        {
            if (previousHeader == null || header.MessageStreamId != previousHeader.MessageStreamId || !header.IsTimerRelative)
                return ChunkMessageHeaderType.New;

            if (header.PacketLength != previousHeader.PacketLength || header.MessageType != previousHeader.MessageType)
                return ChunkMessageHeaderType.SameSource;

            if (header.Timestamp != previousHeader.Timestamp)  
                return ChunkMessageHeaderType.TimestampAdjustment;

            return ChunkMessageHeaderType.Continuation;
        }

        public void Queue(RtmpEvent message, int streamId, int messageStreamId)
        {
            var header = new RtmpHeader();
            var packet = new RtmpPacket(header, message);

            header.StreamId = streamId;
            header.Timestamp = message.Timestamp;
            header.MessageStreamId = messageStreamId;
            header.MessageType = message.MessageType; 
            if (message.Header != null) 
                header.IsTimerRelative = message.Header.IsTimerRelative;
            _queuedPackets.Enqueue(packet);
            Interlocked.Exchange(ref _packetAvailable, 1);
        }

        private void WritePacket(RtmpPacket packet)
        {
            var header = packet.Header;
            var streamId = header.StreamId;
            var message = packet.Body;

            var buffer = GetMessageBytes(header, message);
            header.PacketLength = buffer.Length;

            _rtmpHeaders.TryGetValue(streamId, out var previousHeader);

            _rtmpHeaders[streamId] = header;

            WriteMessageHeader(header, previousHeader);

            var first = true;
            for (var i = 0; i < header.PacketLength; i += _writeChunkSize)
            {
                if (!first)
                    WriteBasicHeader(ChunkMessageHeaderType.Continuation, header.StreamId);

                var bytesToWrite = i + _writeChunkSize > header.PacketLength ? header.PacketLength - i : _writeChunkSize;
                _writer.Write(buffer, i, bytesToWrite);
                first = false;
            }

            if (message is ChunkSize chunkSizeMsg)
                _writeChunkSize = chunkSizeMsg.Size;

        }

        private void WriteBasicHeader(ChunkMessageHeaderType messageHeaderFormat, int streamId)
        {
            var fmt = (byte)messageHeaderFormat;
            if (streamId <= 63)
            {
                _writer.WriteByte((byte)((fmt << 6) + streamId));
            }
            else if (streamId <= 320)
            {
                _writer.WriteByte((byte)(fmt << 6));
                _writer.WriteByte((byte)(streamId - 64));
            }
            else
            {
                _writer.WriteByte((byte)((fmt << 6) | 1));
                _writer.WriteByte((byte)((streamId - 64) & 0xff));
                _writer.WriteByte((byte)((streamId - 64) >> 8));
            }
        }

        private void WriteMessageHeader(RtmpHeader header, RtmpHeader previousHeader)
        {
            var headerType = GetMessageHeaderType(header, previousHeader);
            WriteBasicHeader(headerType, header.StreamId);
            
            var uint24Timestamp = header.Timestamp < 0xFFFFFF ? header.Timestamp : 0xFFFFFF;
            switch (headerType)
            {
                case ChunkMessageHeaderType.New:
                    _writer.WriteUInt24(uint24Timestamp);
                    _writer.WriteUInt24(header.PacketLength);
                    _writer.WriteByte((byte)header.MessageType);
                    _writer.WriteReverseInt(header.MessageStreamId);
                    break;
                case ChunkMessageHeaderType.SameSource:
                    _writer.WriteUInt24(uint24Timestamp);
                    _writer.WriteUInt24(header.PacketLength);
                    _writer.WriteByte((byte)header.MessageType);
                    break;
                case ChunkMessageHeaderType.TimestampAdjustment:
                    _writer.WriteUInt24(uint24Timestamp);
                    break;
                case ChunkMessageHeaderType.Continuation:
                    break;
                default:
                    throw new ArgumentException("headerType");
            }

            if (uint24Timestamp >= 0xFFFFFF)
                _writer.WriteInt32(header.Timestamp);
        }

        private byte[] GetMessageBytes(RtmpEvent message, Action<AmfWriter, RtmpEvent> handler)
        {
            using (var stream = new MemoryStream())
            using (var messageWriter = new AmfWriter(stream, _writer.SerializationContext, _objectEncoding))
            {
                handler(messageWriter, message);
                return stream.ToArray();
            }
        }
        private byte[] GetMessageBytes(RtmpHeader header, RtmpEvent message)
        {
            switch (header.MessageType)
            {
                case MessageType.SetChunkSize:
                    return GetMessageBytes(message, (w, o) => w.WriteInt32(((ChunkSize)o).Size));
                case MessageType.AbortMessage:
                    return GetMessageBytes(message, (w, o) => w.WriteInt32(((Abort)o).StreamId));
                case MessageType.Acknowledgement:
                    return GetMessageBytes(message, (w, o) => w.WriteInt32(((Acknowledgement)o).BytesRead));
                case MessageType.UserControlMessage:
                    return GetMessageBytes(message, (w, o) =>
                    {
                        var m = (UserControlMessage)o;
                        w.WriteUInt16((ushort)m.EventType);
                        foreach (var v in m.Values)
                            w.WriteInt32(v);
                    });
                case MessageType.WindowAcknowledgementSize:
                    return GetMessageBytes(message, (w, o) => w.WriteInt32(((WindowAcknowledgementSize)o).Count));
                case MessageType.SetPeerBandwith:
                    return GetMessageBytes(message, (w, o) =>
                    {
                        var m = (PeerBandwidth)o;
                        w.WriteInt32(m.AcknowledgementWindowSize);
                        w.WriteByte((byte)m.LimitType);
                    });
                case MessageType.Audio:
                case MessageType.Video:
                    return GetMessageBytes(message, (w, o) => WriteData(w, o, ObjectEncoding.Amf0));


                case MessageType.DataAmf0:
                    return GetMessageBytes(message, (w, o) => WriteCommandOrData(w, o, ObjectEncoding.Amf0));
                case MessageType.SharedObjectAmf0:
                    return new byte[0]; // todo: SharedObject
                case MessageType.CommandAmf0:
                    return GetMessageBytes(message, (w, o) => WriteCommandOrData(w, o, ObjectEncoding.Amf0));


                case MessageType.DataAmf3:
                    return GetMessageBytes(message, (w, o) => WriteData(w, o, ObjectEncoding.Amf3));
                case MessageType.SharedObjectAmf3:
                    return new byte[0]; // todo: SharedObject
                case MessageType.CommandAmf3:
                    return GetMessageBytes(message, (w, o) =>
                    {
                        w.WriteByte(0);
                        WriteCommandOrData(w, o, ObjectEncoding.Amf3);
                    });

                case MessageType.Aggregate:
                    // todo: Aggregate messages
                    System.Diagnostics.Debugger.Break();
                    return new byte[0]; // todo: Aggregate
                default:
                    throw new ArgumentOutOfRangeException("Unknown RTMP message type: " + (int)header.MessageType);
            }
        }

        private void WriteData(AmfWriter writer, RtmpEvent o, ObjectEncoding encoding)
        {
            if (o is Command)
                WriteCommandOrData(writer, o, encoding);
            else if (o is ByteData data)
                writer.WriteBytes(data.Data);
        }

       private void WriteCommandOrData(AmfWriter writer, RtmpEvent o, ObjectEncoding encoding)
        {
            var command = o as Command;
            var methodCall = command.MethodCall;
            var isInvoke = command is Invoke;

            // write the method name or result type (first section)
            var isRequest = methodCall.CallStatus == CallStatus.Request;
            if (isRequest)
                writer.WriteAmfItem(encoding, methodCall.Name);
            else
                writer.WriteAmfItem(encoding, methodCall.IsSuccess ? "_result" : "_error");

            if (methodCall.Name == "@setDataFrame")
            {
                writer.WriteAmfItem(encoding, command.ConnectionParameters);
            }

            if (isInvoke)
            {
                writer.WriteAmfItem(encoding, command.InvokeId);
                writer.WriteAmfItem(encoding, command.ConnectionParameters);


                if (!methodCall.IsSuccess)
                    methodCall.Parameters = new object[] { new StatusAsObject(StatusCode.CallFailed, "error", "Call failed.") };

            }
            // write arguments
            foreach (var arg in methodCall.Parameters)
                writer.WriteAmfItem(encoding, arg);
        }

       public void Close()
       {
           _writer?.Dispose();
       }
    }
}
