﻿using System;
using System.Collections.Generic;
using System.Linq;
using WingRtmpService.EmbeddedRtmpServer.Messaging.Events;

namespace WingRtmpService.EmbeddedRtmpServer.Net
{
    class RtmpPublisher: IPublisher
    {        
        private readonly RtmpConnection _connection;
        private readonly StreamRecorder _streamRecorder;
        private readonly List<VideoCacheGroup> _videoCaches = new List<VideoCacheGroup>();

        public event EventHandler<VideoData> VideoDataReceived;

        public event EventHandler<AudioData> AudioDataReceive;

        public event EventHandler VideoMetaDataConfigured;

        public event EventHandler AudioMetaDataConfigured;

        public event EventHandler Closed;

        public string ChannelName { get; }

        public bool IsClosed { get; private set; }

        public VideoData VideoConfigureRecord { get; private set; }

        public AudioData AudioConfigureRecord { get; private set; }

        public RtmpPublisher(string channelName, RtmpConnection connection)
        {
            ChannelName = channelName;
            if (RtmpServerConfig.EnableRecordLiveStream)
            {
                _streamRecorder = new StreamRecorder(this);
            }
            _connection = connection;
            _connection.ConnectionType = ConnectionType.Publisher;
            _connection.UnPublish += (sender, args) => OnClosed(); // will not close the connection.
            _connection.Closed += (sender, args) => OnClosed();
            _connection.ChannelDataReceived += (sender, args) =>
            {
                if (args.Data is VideoData videoData)
                {
                    if (VideoConfigureRecord == null && videoData.Data.Length >= 2 && videoData.Data[1] == 0)
                    {
                        VideoConfigureRecord = videoData;
                        OnVideoMetaDataConfigured();
                    }
                    else
                    {
                        if (RtmpServerConfig.EnablePublishVideoCache)
                        {
                            CacheVideoData(videoData);
                        }
                        OnVideoDataReceived(videoData);
                    }
                }

                if (args.Data is AudioData audioData)
                {
                    if (AudioConfigureRecord == null && audioData.Data.Length >= 2 && audioData.Data[1] == 0)
                    {
                        AudioConfigureRecord = audioData;
                        OnAudioMetaDataConfigured();
                    }
                    else
                    {
                        OnAudioDataReceive(audioData);
                    }
                }
            };
        }

        private void CacheVideoData(VideoData videoData)
        {
            if (videoData.IsKeyFrame)
            {
                lock (_videoCaches)
                {
                    var groupSize = _videoCaches.Count;
                    if (groupSize >= RtmpServerConfig.PublishVideoCacheGroupSize)
                    {
                        _videoCaches.RemoveAt(0);
                    }
                    var cacheGroup = new VideoCacheGroup();
                    cacheGroup.Add(videoData);
                    _videoCaches.Add(cacheGroup);
                }
            }
            else
            {
                lock (_videoCaches)
                {
                    var lastCacheGroup = _videoCaches.LastOrDefault();
                    lastCacheGroup?.Add(videoData);
                }
            }
        }

        public VideoData[] GetVideoCaches()
        {
            var videoCaches = new List<VideoData>();
            lock (_videoCaches)
            {
                foreach (var videoCacheGroup in _videoCaches)
                {
                    foreach (var cache in videoCacheGroup.Cache)
                    {
                        videoCaches.Add(cache);
                    }
                }
            }

            return videoCaches.ToArray();
        }

        public void Close()
        {
            _connection.Close();
            IsClosed = true;
        }

        private void OnVideoDataReceived(VideoData e)
        {
            VideoDataReceived?.Invoke(this, e);
        }

        private void OnAudioDataReceive(AudioData e)
        {
            AudioDataReceive?.Invoke(this, e);
        }

        private void OnClosed()
        {
            if (RtmpServerConfig.EnableRecordLiveStream)
            {
                _streamRecorder?.Dispose();
            }
            Closed?.Invoke(this, EventArgs.Empty);
        }

        private void OnVideoMetaDataConfigured()
        {
            VideoMetaDataConfigured?.Invoke(this, EventArgs.Empty);
        }

        private void OnAudioMetaDataConfigured()
        {
            AudioMetaDataConfigured?.Invoke(this, EventArgs.Empty);
        }
    }
}
