﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;
using WingRtmpService.EmbeddedRtmpServer.IO;
using WingRtmpService.EmbeddedRtmpServer.Log;
using WingRtmpService.EmbeddedRtmpServer.Messaging.Events;
using WingRtmpService.EmbeddedRtmpServer.Net.Http;

namespace WingRtmpService.EmbeddedRtmpServer.Net
{
    public class RtmpServer
    {
        #region handshake

        const int HandshakeRandomSize = 1528;

        // size for c0, c1, s1, s2 packets. c0 and s0 are 1 byte each.
        const int HandshakeSize = HandshakeRandomSize + 4 + 4;

        public struct Handshake
        {
            // C0/S0 only
            public byte Version;

            // C1/S1/C2/S2
            public uint Time;
            // in C1/S1, MUST be zero. in C2/S2, time at which C1/S1 was read.
            public uint Time2;

            public byte[] Random;

            public Handshake Clone()
            {
                return new Handshake()
                {
                    Version = Version,
                    Time = Time,
                    Time2 = Time2,
                    Random = Random
                };
            }

            public static Handshake Read(Stream stream, bool readVersion)
            {
                var size = HandshakeSize + (readVersion ? 1 : 0);
                var buffer = StreamHelper.ReadBytes(stream, size);

                using (var reader = new AmfReader(new MemoryStream(buffer), null))
                {
                    return new Handshake()
                    {
                        Version = readVersion ? reader.ReadByte() : default,
                        Time = reader.ReadUInt32(),
                        Time2 = reader.ReadUInt32(),
                        Random = reader.ReadBytes(HandshakeRandomSize)
                    };
                }
            }

            public static void Write(Stream stream, Handshake h, bool writeVersion)
            {
                using (var memoryStream = new MemoryStream())
                using (var writer = new AmfWriter(memoryStream, null))
                {
                    if (writeVersion)
                        writer.WriteByte(h.Version);

                    writer.WriteUInt32(h.Time);
                    writer.WriteUInt32(h.Time2);
                    writer.WriteBytes(h.Random);

                    var buffer = memoryStream.ToArray();
                    stream.Write(buffer, 0, buffer.Length);
                }
            }
        }

        #endregion

        private const int ReceiveTimeout = 10000;
        private const int SendTimeout = 10000;

        private readonly List<string> _registeredApps = new List<string>();

        private readonly ConcurrentDictionary<string, IPublisher> _publishers = new ConcurrentDictionary<string, IPublisher>();
        private readonly ConcurrentDictionary<string, List<IPlayer>> _players = new ConcurrentDictionary<string, List<IPlayer>>();
        private readonly List<RtmpConnection> _connections = new List<RtmpConnection>();
        

        private readonly Random _random = new Random();

        private readonly TcpListener _listener;

        private Thread _ioThread;
        private Thread _listenThread;

        private readonly IHttpExtension _httpExtension;

        public bool Started { get; private set; }

        public RtmpServer()
        {
            if (RtmpServerConfig.EnableHttpFlvExtension || RtmpServerConfig.EnableHlsExtension)
            {
                try
                {
                    _httpExtension = new HttpServer();
                }
                catch (Exception ex)
                {
                    Logger.WriteLineError(ex.Message);
                }

                if (_httpExtension == null)
                {
                    Logger.WriteLineWarn("The http server was not started.");
                }
            }
            _listener = new TcpListener(IPAddress.Any, RtmpServerConfig.RtmpPort);
            RegisterApp(RtmpServerConfig.DefaultApp);
        }

        public void Start()
        {
            _listener.Start();
            Started = true;
            if (RtmpServerConfig.EnableGlobalPollingMode)
            {
                _ioThread = new Thread(() =>
                {
                    //Use 1 thread or 1/2 cpu cores threads to handle the IO.
                    var parallelOptions = new ParallelOptions { MaxDegreeOfParallelism = Math.Max(1, Environment.ProcessorCount / 2) };
                    Logger.WriteLineInfo($"Set parallel tasks:{parallelOptions.MaxDegreeOfParallelism} for IO handler.");
                    while (Started)
                    {
                        RtmpConnection[] connections;
                        lock (_connections)
                        {
                            connections = _connections.ToArray();
                        }

                        if (connections.Length > 0)
                        {
                            Parallel.ForEach(connections, parallelOptions, connection =>
                            {
                                if (!connection.IsClosed)
                                {
                                    try
                                    {
                                        if (!connection.IsConnected)
                                        {
                                            Logger.WriteLineWarn($"[{connection.ConnectionType}] Connection already closed.");
                                            connection.Close();
                                        }
                                        else
                                        {
                                            var hw = connection.WriteOnce();
                                            var hr = connection.ReadOnce();
                                            if (hw || hr)
                                            {
                                                connection.LastPing = DateTime.UtcNow;
                                            }
                                            else
                                            {
                                                if (RtmpServerConfig.AutoCleanNoOperationConnections)
                                                {
                                                    if (DateTime.UtcNow - connection.LastPing >
                                                        RtmpConnection.IoTimeout)
                                                    {
                                                        Logger.WriteLineWarn(
                                                            $"[{connection.ConnectionType}]Connection has no writing or reading over {RtmpConnection.IoTimeout.Seconds} seconds, it will be closed.");
                                                        connection.Close();
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    catch
                                    {
                                        connection.Close();
                                    }
                                }
                            });
                        }

                        Thread.Sleep(1);
                    }
                });
                _ioThread.Start();
                Logger.WriteLineDebug("Global polling thread started.");
            }
            _listenThread = new Thread(() =>
            {
                try
                {
                    while (Started)
                    {
                        var client = _listener.AcceptTcpClient();
                        client.ReceiveTimeout = ReceiveTimeout;
                        client.SendTimeout = SendTimeout;
                        ThreadPool.QueueUserWorkItem(state =>
                        {
                            try
                            {
                                DoHandshake(client);
                            }
                            catch
                            {
                                client.Close();
                            }
                        });
                    }
                }
                catch (Exception e)
                {
                    Logger.WriteLineError(e.ToString());
                }
            });
            _listenThread.Start();
            Logger.WriteLineInfo("RTMP server started.");
        }


        public void Stop()
        {
            Started = false;
            if (RtmpServerConfig.EnableHttpFlvExtension || RtmpServerConfig.EnableHlsExtension)
            {
                _httpExtension?.Close();
            }

            if (RtmpServerConfig.EnableGlobalPollingMode)
            {
                _ioThread.Join();
            }
            _listenThread.Join();

            lock (_connections)
            {
                foreach (var connection in _connections.ToArray())
                {
                    try
                    {
                        connection.Close();
                    }
                    catch (Exception e)
                    {
                        Logger.WriteLineError("Error when close client during stop server: " + e.Message);
                    }
                }
            }

            _listener.Stop();
        }


        public void DoHandshake(TcpClient tcpClient)
        {
            Stream stream = tcpClient.GetStream();
            var randomBytes = new byte[1528];
            _random.NextBytes(randomBytes);
            var s01 = new Handshake()
            {
                Version = 3,
                Time = (uint)Environment.TickCount,
                Time2 = 0,
                Random = randomBytes
            };

            var c01 = Handshake.Read(stream, true);
            //S0 + S1
            Handshake.Write(stream, s01, true);
            //S0 + S1 + S2
            //S2 = S1
            var s2 = c01.Clone();
            Handshake.Write(stream, s2, false);
            //Read C2
            var c2 = Handshake.Read(stream, false);

            //handshake check
            if (!c2.Random.SequenceEqual(s01.Random))
                throw new ProtocolViolationException();

            var connection = new RtmpConnection(tcpClient);
            connection.RequestConnect += OnConnectionRequestConnect;
            connection.RequestPublish += OnConnectionRequestPublish;
            connection.Published += OnConnectionPublished;
            connection.RequestPlay += OnConnectionRequestPlay;
            connection.Played += OnConnectionPlayed;
            connection.Closed += OnConnectionClosed;
            lock (_connections)
            {
                _connections.Add(connection);
            }
        }

        #region Connection
        private void OnConnectionRequestConnect(object sender, ConnectEventArgs e)
        {
            lock (_registeredApps)
            {
                e.AuthSuccess = _registeredApps.IndexOf(e.App) != -1;
                if (!e.AuthSuccess)
                {
                    Logger.WriteLineError($"Auth connect failed, connect app is {e.App}, exist apps are [{string.Join(",", _registeredApps)}]");
                }
            }
        }

        private void OnConnectionClosed(object sender, EventArgs e)
        {
            var connection = (RtmpConnection)sender;
            connection.RequestConnect -= OnConnectionRequestConnect;
            connection.RequestPublish -= OnConnectionRequestPublish;
            connection.Published -= OnConnectionPublished;
            connection.RequestPlay -= OnConnectionRequestPlay;
            connection.Played -= OnConnectionPlayed;
            connection.Closed -= OnConnectionClosed;
            lock (_connections)
            {
                _connections.Remove(connection);
            }
        }
        #endregion

        #region Publisher

        private void OnConnectionRequestPublish(object sender, PublishEventArgs e)
        {
            var channelName = e.App + "/" + e.Path;
            if (_publishers.ContainsKey(channelName))
            {
                Logger.WriteLineWarn($"Publisher for channel:{channelName} already exists.");
                e.CanPublish = false;
            }
            else
            {
                e.CanPublish = true;
            }
        }

        private void OnConnectionPublished(object sender, PublishEventArgs e)
        {
            var connection = (RtmpConnection)sender;
            var channelName = e.App + "/" + e.Path;
            var publisher = new RtmpPublisher(channelName, connection);
            publisher.Closed += OnPublisherClosed;
            publisher.VideoMetaDataConfigured += OnPublisherVideoMetaDataConfigured;
            publisher.AudioMetaDataConfigured += OnPublisherAudioMetaDataConfigured;
            publisher.VideoDataReceived += OnPublisherVideoDataReceived;
            publisher.AudioDataReceive += OnPublisherAudioDataReceive;
            _publishers.AddOrUpdate(channelName, cn => publisher,
                (cn, exist) =>
                {
                    exist.Closed -= OnPublisherClosed;
                    exist.Close();
                    return publisher;
                });
            if (RtmpServerConfig.EnableHttpFlvExtension || RtmpServerConfig.EnableHlsExtension)
            {
                _httpExtension?.AddPublisher(publisher);
            }
        }

        private void OnPublisherAudioMetaDataConfigured(object sender, EventArgs e)
        {
            var publisher = (IPublisher)sender;
            if (_players.TryGetValue(publisher.ChannelName, out var players))
            {
                lock (players)
                {
                    foreach (var player in players)
                    {
                        try
                        {
                            player.SendMetaData(null, publisher.AudioConfigureRecord);
                        }
                        catch (Exception ex)
                        {
                            Logger.WriteLineWarn($"Send audio meta data to player failed: {ex.Message}");
                        }
                    }
                }
            }
        }

        private void OnPublisherVideoMetaDataConfigured(object sender, EventArgs e)
        {
            var publisher = (IPublisher)sender;
            if (_players.TryGetValue(publisher.ChannelName, out var players))
            {
                lock (players)
                {
                    foreach (var player in players)
                    {
                        try
                        {
                            player.SendMetaData(publisher.VideoConfigureRecord, null);
                        }
                        catch (Exception ex)
                        {
                            Logger.WriteLineWarn($"Send video meta data to player failed: {ex.Message}");
                        }
                    }
                }
            }
        }


        private void OnPublisherAudioDataReceive(object sender, AudioData e)
        {
            var publisher = (IPublisher)sender;
            if (_players.TryGetValue(publisher.ChannelName, out var players))
            {
                lock (players)
                {
                    foreach (var player in players)
                    {
                        try
                        {
                            player.SendAudioData(e);
                        }
                        catch (Exception ex)
                        {
                            Logger.WriteLineWarn($"Send audio data to player failed: {ex.Message}");
                        }
                    }
                }
            }
        }

        private void OnPublisherVideoDataReceived(object sender, VideoData e)
        {
            var publisher = (IPublisher)sender;
            if (_players.TryGetValue(publisher.ChannelName, out var players))
            {
                lock (players)
                {
                    foreach (var player in players)
                    {
                        try
                        {
                            player.SendVideoData(e);
                        }
                        catch (Exception ex)
                        {
                            Logger.WriteLineWarn($"Send video data to player failed: {ex.Message}");
                        }
                    }
                }
            }
        }

        private void OnPublisherClosed(object sender, EventArgs e)
        {
            var publisher = (IPublisher)sender;
            publisher.Closed -= OnPublisherClosed;
            publisher.VideoMetaDataConfigured -= OnPublisherVideoMetaDataConfigured;
            publisher.AudioMetaDataConfigured -= OnPublisherAudioMetaDataConfigured;
            publisher.VideoDataReceived -= OnPublisherVideoDataReceived;
            publisher.AudioDataReceive -= OnPublisherAudioDataReceive;
            if (_publishers.TryRemove(publisher.ChannelName, out _))
            {
                publisher.Close();
                if (_players.TryRemove(publisher.ChannelName, out var players))
                {
                    lock (players)
                    {
                        foreach (var player in players)
                        {
                            player.Close();
                        }
                        players.Clear();
                    }
                }
                else
                {
                    Logger.WriteLineWarn($"No players for channel:{publisher.ChannelName}");
                }
            }
            else
            {
                Logger.WriteLineWarn($"No publisher for channel:{publisher.ChannelName}");
            }
        }

        #endregion

        #region player
        private void OnConnectionRequestPlay(object sender, PlayEventArgs e)
        {
            if (!RtmpServerConfig.AllowPlayWithoutPublish)
            {
                var channelName = e.App + "/" + e.Path;
                if (_publishers.ContainsKey(channelName))
                {
                    e.CanPlay = true;
                }
                else
                {
                    Logger.WriteLineWarn($"No publisher for channel:{channelName}.");
                    e.CanPlay = false;
                }
            }
            else
            {
                e.CanPlay = true;
            }
   
        }

        private void OnConnectionPlayed(object sender, PlayEventArgs e)
        {
            var connection = (RtmpConnection)sender;
            var channelName = e.App + "/" + e.Path;
            var player = new RtmpPlayer(channelName, connection);
            player.Closed += OnPlayerClosed;

            if (_publishers.TryGetValue(player.ChannelName, out var publisher))
            {
                //Send meta data
                player.SendMetaData(publisher.VideoConfigureRecord, publisher.AudioConfigureRecord);
                if (RtmpServerConfig.EnablePublishVideoCache)
                {
                    //Fast play
                    var videoCaches = publisher.GetVideoCaches();
                    foreach (var videoCache in videoCaches)
                    {
                        player.SendVideoData(videoCache);
                    }
                }
            }
            else
            {
                if (!RtmpServerConfig.AllowPlayWithoutPublish)
                {
                    Logger.WriteLineWarn($"Request publisher for channel {player.ChannelName} does not exist.");
                    player.Close();
                }
            }

            if (!player.IsClosed)
            {
                _players.AddOrUpdate(channelName, cn =>
                {
                    var players = new List<IPlayer> { player };
                    return players;
                }, (cn, players) =>
                {
                    lock (players)
                    {
                        players.Add(player);
                        return players;
                    }
                });
            }
        }

        private void OnPlayerClosed(object sender, EventArgs e)
        {
            var player = (IPlayer)sender;
            player.Closed -= OnPlayerClosed;
            if (_players.TryGetValue(player.ChannelName, out var players))
            {
                lock (players)
                {
                    players.Remove(player);
                    if (players.Count == 0)
                    {
                        _players.TryRemove(player.ChannelName, out _);
                    }
                    else
                    {
                        _players.AddOrUpdate(player.ChannelName, cn => players, (cn, exist) => players);
                    }
                }
            }
            Logger.WriteLineInfo($"One player for channel:{player.ChannelName} closed.");

        }

        #endregion

        public void RegisterApp(string appName)
        {
            lock (_registeredApps)
            {
                if (_registeredApps.IndexOf(appName) != -1)
                {
                    throw new InvalidOperationException("app exists");
                }
                _registeredApps.Add(appName);
            }
        }
    }
}
