﻿using System;
using System.Collections.Concurrent;
using System.IO;
using System.Threading;
using WingRtmpService.EmbeddedRtmpServer.Log;
using WingRtmpService.EmbeddedRtmpServer.Messaging.Events;

namespace WingRtmpService.EmbeddedRtmpServer.Net
{
    class StreamRecorder:IDisposable
    {
        private readonly IPublisher _publisher;
        private readonly Stream _recordStream;
        private readonly BinaryWriter _recordWriter;
        private readonly Thread _recordThread;
        private readonly ConcurrentQueue<ByteData> _recordDataQueue = new ConcurrentQueue<ByteData>();
        private int _dataAvailable;

        private bool _closed;
        private bool _disposed;

        public StreamRecorder(IPublisher publisher)
        {
            _publisher = publisher;
            var sections = publisher.ChannelName.Split('/');
            if (sections.Length == 2)
            {
                var app = sections[0];
                var stream = sections[1];
                var now = DateTime.Now;
                if (!Directory.Exists(RtmpServerConfig.RecordFolder))
                {
                    Directory.CreateDirectory(RtmpServerConfig.RecordFolder);
                }

                var recordFolder = Path.Combine(RtmpServerConfig.RecordFolder, app);
                if (!Directory.Exists(recordFolder))
                {
                    Directory.CreateDirectory(recordFolder);
                }

                var recordFile = Path.Combine(recordFolder, $"{stream}_{now:yyyyMMddHHmmss}{now.Millisecond:d3}.flv");
                _recordStream = File.Create(recordFile);
                _recordWriter = new BinaryWriter(_recordStream);
                _recordWriter.Write(FlvTag.FlvHeader);
                _publisher.AudioMetaDataConfigured += OnAudioMetaDataConfigured;
                _publisher.VideoMetaDataConfigured += OnVideoMetaDataConfigured;
                _publisher.AudioDataReceive += OnAudioDataReceive;
                _publisher.VideoDataReceived += OnVideoDataReceived;
                _recordThread = new Thread(() =>
                {
                    while (!_closed)
                    {
                        WriteOnce();
                        Thread.Sleep(1);
                    }
                });
                _recordThread.Start();
                Logger.WriteLineInfo($"Stream recorder for channel:{_publisher.ChannelName} started.");
            }
            else
            {
                throw new InvalidOperationException("Can not parser the channel name.");
            }
        }

        ~StreamRecorder()
        {
            DoDispose();
        }

        private void OnVideoDataReceived(object sender, VideoData e)
        {
            Record(e);
        }

        private void OnAudioDataReceive(object sender, AudioData e)
        {
            Record(e);
        }

        private void OnVideoMetaDataConfigured(object sender, EventArgs e)
        {
            Record(_publisher.VideoConfigureRecord);
        }

        private void OnAudioMetaDataConfigured(object sender, EventArgs e)
        {
            Record(_publisher.AudioConfigureRecord);
        }

        private void Record(ByteData data)
        {
            _recordDataQueue.Enqueue(data);
            Interlocked.Exchange(ref _dataAvailable, 1);
        }

        private void WriteOnce()
        {
            if (Interlocked.Exchange(ref _dataAvailable, 0) == 1)
            {
                while (_recordDataQueue.TryDequeue(out var data))
                {
                    var flvTag = new FlvTag(data);
                    _recordWriter.Write(flvTag.TagData);
                }
            }
        }

        private void DoDispose()
        {
            if (!_disposed)
            {
                _publisher.AudioMetaDataConfigured -= OnAudioMetaDataConfigured;
                _publisher.VideoMetaDataConfigured -= OnVideoMetaDataConfigured;
                _publisher.AudioDataReceive -= OnAudioDataReceive;
                _publisher.VideoDataReceived -= OnVideoDataReceived;
                _closed = true;
                _recordThread.Join();
                _recordWriter.Flush();
                _recordWriter?.Dispose();
                _recordStream?.Dispose();
                Logger.WriteLineInfo($"Stream recorder for channel:{_publisher.ChannelName} closed.");
                _disposed = true;
            }
        }

        public void Dispose()
        {
            DoDispose();
            GC.SuppressFinalize(this);
        }
    }
}
